from django.db import models
from django.db.models.sql import Query

from raw_sugar.sources import SourceRaw, FromQuerySet


class RawSugarQuery(Query):
    def __init__(self, *args,
                 _source_func=None,
                 _source_func_args=None,
                 _source_func_kwargs=None,
                 **kwargs):
        self._source_func = _source_func
        self._source_func_args = _source_func_args
        self._source_func_kwargs = _source_func_kwargs
        return super().__init__(*args, **kwargs)

    def get_compiler(self, *args, **kwargs):
        compiler = super().get_compiler(*args, **kwargs)
        get_from_clause_method = compiler.get_from_clause

        def get_from_clause_wrapper(*args, **kwargs):
            source = self._source_func(
                self.model,
                *(self._source_func_args or []),
                **(self._source_func_kwargs or {}))
            assert isinstance(source, SourceRaw)
            if isinstance(source, FromQuerySet):
                source._set_target_model(self.model)

            qn = compiler.connection.ops.quote_name
            result, params = get_from_clause_method(*args, **kwargs)
            assert result[0] == qn(self.model._meta.db_table)
            if len(source.translations) > 0 or len(source.null_fields) > 0:
                select_fields = []
                for col in self.model._meta.fields:
                    field_name = col.column
                    if not field_name in source.translations.values()\
                            and not field_name in source.null_fields:
                        select_fields.append('{}.{}'.format(
                            qn('wrapper_table'),
                            qn(field_name)))
                for null_field in source.null_fields:
                    select_fields.append('NULL AS {}'.format(
                        qn(null_field)))
                for translation_field in source.translations.items():
                    select_fields.append('{}.{} AS {}'.format(
                        qn('wrapper_table'),
                        qn(translation_field[0]),
                        qn(translation_field[1])))
                wrapper = '(SELECT {} FROM {} AS {})'.format(
                    ', '.join(select_fields),
                    source.raw_query,
                    qn('wrapper_table'))
            else:
                wrapper = source.raw_query
            result[0] = '{} AS {}'.format(
                wrapper, qn(self.model._meta.db_table))
            params = tuple(source.params) + tuple(params)
            return result, params

        compiler.get_from_clause = get_from_clause_wrapper
        return compiler


class RawSugarQuerySet(models.QuerySet):
    def __init__(self, *args, query=None, _source_func=None,
                 _source_func_args=None, _source_func_kwargs=None,
                 **kwargs):
        empty_query = query is None
        r = super().__init__(*args, query=query, **kwargs)
        if empty_query:
            self.query = RawSugarQuery(
                self.model,
                _source_func=_source_func,
                _source_func_args=_source_func_args,
                _source_func_kwargs=_source_func_kwargs)
        return r

    def with_params(self, *args, **kwargs):
        clone = self._chain()
        clone.query._source_func_args = args
        clone.query._source_func_kwargs = kwargs
        return clone



from django.db import models
from .query import RawSugarQuerySet
from .sources import SourceRaw, FromRaw, FromQuerySet


class RawManagerMixin:
    def from_raw(self, raw_query=None, params=[],
                 translations={}, null_fields=[],
                 db_table=None):
        def source_func(cls, *args):
            return FromRaw(raw_query, args, translations,
                           null_fields, db_table)
        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source_func=source_func,
                                _source_func_args=params)

    def from_queryset(self, queryset, translations={}):
        def source_func(cls):
            return FromQuerySet(queryset, translations)
        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source_func=source_func)


class RawManager(models.Manager, RawManagerMixin):
    def __init__(self, from_raw=None):
        if from_raw is not None:
            assert isinstance(from_raw, FromRaw), \
                "Expected a `FromRaw` to be passed "\
                "to 'from_raw', but received a `%s" % type(from_raw)
        self._source = from_raw
        return super().__init__()

    def get_queryset(self):
        if isinstance(self._source, FromRaw):
            def source_func(cls, *args):
                return self._source.with_params(*args)
            return RawSugarQuerySet(self.model,
                                    using=self._db,
                                    _source_func=source_func,
                                    _source_func_args=self._source.params)
        else:
            return super().get_queryset()

    def with_params(self, *args):
        def source_func(cls, *args):
            return self._source.with_params(*args)
        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source_func=source_func,
                                _source_func_args=args)


class DecoratedRawManager(models.Manager, RawManagerMixin):
    def __init__(self, *args, _source_func=None, _source_func_is_callable=False,
                 _set_source_func_on_next_call=False, **kwargs):
        self._source_func = _source_func
        if not _source_func_is_callable:
            assert callable(self._source_func)
        self._source_func_is_callable = _source_func_is_callable
        self._set_source_func_on_next_call = _set_source_func_on_next_call
        return super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self._set_source_func_on_next_call:
            self._source_func = args[0]
            assert callable(self._source_func)
            self._set_source_func_on_next_call = False
            return self
        elif self._source_func_is_callable:
            return RawSugarQuerySet(
                self.model, using=self._db,
                _source_func=self._source_func,
                _source_func_args=args,
                _source_func_kwargs=kwargs)
        else:
            raise TypeError

    def get_queryset(self):
        return RawSugarQuerySet(
            self.model, using=self._db,
            _source_func=self._source_func)


def raw_manager(is_callable=False):
    if callable(is_callable):
        return DecoratedRawManager(_source_func=is_callable)
    if is_callable:
        return DecoratedRawManager(_set_source_func_on_next_call=True,
                                   _source_func_is_callable=True)
    else:
        return DecoratedRawManager(_set_source_func_on_next_call=True)



import os
from setuptools import setup, find_packages

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

with open("README.md", 'r') as f:
    long_description = f.read()
    print(long_description)

setup(
    name='django-raw-sugar',
    version='0.1.16',
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    description='Turns your raw sql into a QuerySet',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zxibizz/django-raw-sugar',
    author='Roman Lee',
    author_email='romanlee1996@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from django.db import models
from raw_sugar import raw_manager, RawManager, FromRaw


class AnotherSimpleModel(models.Model):
    pass


class MySimpleModel(models.Model):
    name = models.TextField()
    number = models.IntegerField()
    source = models.ForeignKey(
        AnotherSimpleModel, models.DO_NOTHING, null=True)

    objects = RawManager()

    my_raw_manager = RawManager(FromRaw('SELECT "my str" as name, 111 as number',
                                        null_fields=['id', 'source_id']))

    @raw_manager
    def my_raw_manager_2(cls):
        return FromRaw('SELECT "my str" as name, 111 as number',
                       null_fields=['id', 'source_id'])

    @raw_manager(is_callable=True)
    def my_callable_raw_manager(cls, name):
        return FromRaw('SELECT %s as name, 111 as number',
                       null_fields=['id', 'source_id'],
                       params=[name])
